(* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *)
(*  File: CROPBMPE.PAS                                                     *)
(*  Contents: program for cropping 24 bpp BMPs and converting them to 4 bpp*)
(*  Author: Mircosoft (http://mircosoft.mzf.cz)                            *)
(*  Last modified: ?.?.2011                                                *)
(*  To compile: (DOS.TPU)                                                  *)
(*  To run: nothing                                                        *)
(*  Warning: you use these source codes at your own risk                   *)
(* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *)
program BMP_cropper;
{$I+}
uses nfsup,paleta2,errmsg,dos,retezce;

type polebytu=array[0..767] of byte; {pomocny typ pro pristup k jednotlivym bytum bitmap
                 (rozsah 768 B je zvolen jenom proto, aby slo pole pretypovat
                  na takhle velkou paletu, ale obecne na nem vubec nezalezi)}

     HlavickaBMP = record
                   znacka:array[1..2] of char; {'BM'}
                   VelikostSouboru:longint; {uplne celeho}
                   rezervovano:longint; {k nicemu}
                   OfsetBitmapy:longint; {adresa prvniho bytu bitmapy, poc. od 0 od zacatku souboru}
                   end;

     HlavickaDIBv3 = record {od Windows 3.0 nahoru, nejpouzivanejsi}
                     VelikostHlavicky:longint; {=40}
                     _sirka, {sirka bitmapy [px]}
                     _vyska:longint; {vyska [px]}
                     PocetBitovychRovin:word; {=1 (vzdy)}
                     BituNaPixel:word; {=1, 4, 8, 16, 24 nebo 32}
                     komprese:longint; {0=zadna, 1=RLE pro 8 b/px, 2=RLE pro 4 b/px, 3=bitove pole (?), 4=JPG, 5=PNG}
                     VelikostBitmapy:longint; {=vyska * zaokrouhlena sirka (nahoru na nasobek 4)}
                     rozliseniX,rozliseniY:longint; {[px/m]}
                     PocetBarevVPalete:longint; {pocet polozek palety, 0 = vsechny podle bpp}
                     PocetPlatnychBarev:longint; {kolik barev je opravdu dulezitych, 0 = vsechny}
                     end;

var sr:searchrec;
    zadanejmeno:pathstr;
    cesta:dirstr; jmeno:namestr; koncovka:extstr;
    vsechny:boolean;
    buffer1,buffer2:^polebytu; {univerzalni dynamicky odkladaci prostor}
    l:longint;
    r,g,b,pixel,bpp:byte;
    x,y,delkaradku1,delkaradku2:word;
    i,barva:integer;
    NaslaSe:boolean;
    VstupniIndex,VystupniIndex:word;
    hbmp:hlavickabmp;
    hdib3:hlavickadibv3;
    sirka,vyska:longint;
    vstup,vystup:file;
    radek:pointer;
    paleta:array[0..15,1..4] of byte; {B,G,R,P}
    jmenovstupu,jmenovystupu:pathstr;
    vlevo,vpravo,nahore,dole:word;
    nastaveni:text;

BEGIN
writeln('--- Automatic BMP cropper ---');
writeln;
if paramcount=0 then begin
                     writeln('Usage:');
                     writeln('1) CROP image.bmp');
                     writeln('   Reads the given image, saves its cropped version under the same name');
                     writeln('   and backups the original one as image.BAK.');
                     writeln;
                     writeln('2) CROP /ALL image.bmp');
                     writeln('   Does the same with all files in the directory where the given image');
                     writeln('   is located (regardless of which of them you put in the parameter, only');
                     writeln('   the path is extracted from it).');
                     writeln;
                     writeln('Cropping dimensions can be set in file CROPBMPE.CFG.');
                     writeln('(Only change the numbers, don''t move the lines!)');
                     writeln;
                     writeln('File names should be max. 8 characters long, otherwise they will be');
                     writeln('shortened (for example, "VERY LONG NAME.BMP" to "VERYLO~1.BMP").');
                     writeln;
                     write('Press Enter to quit...');
                     readln;
                     halt;
                     end;
if stringup(paramstr(1))='/ALL' then begin
                                     vsechny:=true;
                                     zadanejmeno:=paramstr(2);
                                     end
                                else begin
                                     vsechny:=false;
                                     zadanejmeno:=paramstr(1);
                                     end;
if zadanejmeno='' then chyba('Missing input file, cannot determine directory to process.',0);

if vsechny then begin
                fsplit(zadanejmeno,cesta,jmeno,koncovka);
                findfirst(cesta+'*.BMP',anyfile-directory-readonly,sr);
                if doserror<>0 then chyba('No BMP files in this directory. Make sure the working folder is set correctly.',0);
                end;

{$I-}
assign(nastaveni,'CROPBMPE.CFG');
reset(nastaveni);
readln(nastaveni);
readln(nastaveni,vlevo);
readln(nastaveni);
readln(nastaveni,vpravo);
readln(nastaveni);
readln(nastaveni,nahore);
readln(nastaveni);
readln(nastaveni,dole);
close(nastaveni);
if ioresult<>0 then chyba('Configuration file CROPBMPE.CFG not found in the program''s directory.',0);
{$I+}

writeln('Working...');

repeat
if vsechny then jmenovstupu:=cesta+sr.name else jmenovstupu:=zadanejmeno;
assign(vstup,jmenovstupu); reset(vstup,1);
assign(vystup,'OUTPUT.$$$'); rewrite(vystup,1);
blockread(vstup,hbmp,sizeof(hlavickabmp));
blockread(vstup,hdib3.velikosthlavicky,4); {prvni 4 B = velikost hlavicky DIB}
blockread(vstup,hdib3._sirka,hdib3.velikosthlavicky-4); {zbytek hlavicky DIB}
if hdib3.velikosthlavicky=12
  then chyba('DIB header version 1 is not supported.',0)
  else with hdib3 do begin {verze 3}
                     if komprese<>0
                       then chyba('Compressed BMPs not supported.',0)
                       else begin sirka:=_sirka; vyska:=_vyska; end;
                     bpp:=bitunapixel;
                     if bpp=24 then y:=255
                               else chyba(nastr(bpp)+' bpp not supported, save the image again as 24 bits per pixel.',0);
                     end;
{buffer pro vstupni soubor:}
delkaradku1:=(sirka*24) shr 3 + ord((sirka*24) and 7<>0);
while (delkaradku1 and 3)<>0 do inc(delkaradku1); {zaokrouhleni nahoru na nasobek 4}
if (delkaradku1>maxavail) then chyba('Not enough memory to allocate input buffer.',0)
                          else getmem(buffer1,delkaradku1);
{buffer pro vystupni soubor:}
delkaradku2:=((sirka-vlevo-vpravo)*4) shr 3 + ord((sirka*4) and 7<>0);
while (delkaradku2 and 3)<>0 do inc(delkaradku2); {zaokrouhleni nahoru na nasobek 4}
if (2*delkaradku2>maxavail) then chyba('Not enough memory to allocate output buffer.',0)
                            else getmem(buffer2,2*delkaradku2);
{ulozeni hlavicek:}
with hbmp do begin
             znacka:='BM';
             rezervovano:=0;
             VelikostSouboru:=sizeof(hbmp)+delkaradku2*vyska; {zatim}
             OfsetBitmapy:=sizeof(hbmp); {zatim}
             end;
with hdib3 do begin
              VelikostHlavicky:=40;
              _sirka:=sirka-vlevo-vpravo;
              _vyska:=vyska-nahore-dole;
              PocetBitovychRovin:=1;
              BituNaPixel:=4;
              komprese:=0;
              VelikostBitmapy:=vyska*delkaradku2;
              rozliseniX:=0;{7874;}
              rozliseniY:=0;{7874;}
              PocetBarevVPalete:=16;
              PocetPlatnychBarev:=16;
              end;
with hbmp do begin
             inc(VelikostSouboru,sizeof(hdib3)+16 shl 2);
             inc(OfsetBitmapy,sizeof(hdib3)+16 shl 2);
             end;
blockwrite(vystup,hbmp,sizeof(hbmp));
blockwrite(vystup,hdib3,sizeof(hdib3));
seek(vystup,hbmp.ofsetbitmapy); {paletu zatim preskocime}
fillchar32(paleta,sizeof(paleta),0);
{vstup i vystup jsou ted na zacatku bitmapy}
i:=-1; {pocitadlo barev, -1=zatim zadna}
for y:=vyska downto 1 do
 begin
 blockread(vstup,buffer1^,delkaradku1);
 if (y>vyska-dole)or(y<=nahore) then continue; {tyhle vynechame cele}
 for x:=0 to hdib3._sirka-1 do
  begin
  b:=buffer1^[3*(x+vlevo)];
  g:=buffer1^[3*(x+vlevo)+1];
  r:=buffer1^[3*(x+vlevo)+2];
  naslase:=false;
  for barva:=0 to i do if (b=paleta[barva,1])and(g=paleta[barva,2])and(r=paleta[barva,3])
                         then begin
                              naslase:=true;
                              break;
                              end;
  if not naslase then begin
                      inc(i);
                      if i>15 then chyba('Too many unique colours. Save only the terminal window.',0);
                      paleta[i,1]:=b; paleta[i,2]:=g; paleta[i,3]:=r;
                      barva:=i;
                      end;
  buffer2^[x]:=barva;
  end;
 {uprava vystupu na 4 bpp:}
 asm
 push DS
 les DI,buffer2
 mov CX,word ptr hdib3._sirka
 lds SI,buffer2
  @cyklus:
  lodsb
  mov BL,AL
  shl BL,4
  lodsb
  and AL,$0F
  or AL,BL
  stosb
  dec CX
  jz @konec
  dec CX
  jnz @cyklus
 @konec:
 pop DS
 end;
 blockwrite(vystup,buffer2^,delkaradku2);
 end;
{bitmapa je hotova, ted jeste paletu:}
seek(vystup,sizeof(hbmp)+sizeof(hdib3));
blockwrite(vystup,paleta,sizeof(paleta));
{hotovo, uklid:}
close(vstup); close(vystup);
freemem(buffer1,delkaradku1);
freemem(buffer2,2*delkaradku2);
dec(jmenovstupu[0],3);
rename(vstup,jmenovstupu+'BAK'); {zaloha stareho}
rename(vystup,jmenovstupu+'BMP'); {finalni jmeno pro novy}
if vsechny then findnext(sr);
until not vsechny or (doserror<>0);
writeln('...done.');
END.